{
  "cells": [
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "%matplotlib inline"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "\nPynest microcircuit network\n---------------------------\n\nMain file for the microcircuit.\n\n\nThis example uses the function GetNodes, which is deprecated. A deprecation\nwarning is therefore issued. For details about deprecated functions, see\ndocumentation.\n\nAuthors\n~~~~~~~~\n\nHendrik Rothe, Hannah Bos, Sacha van Albada; May 2016\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "import nest\nimport numpy as np\nimport os\nfrom helpers import adj_w_ext_to_K\nfrom helpers import synapses_th_matrix\nfrom helpers import get_total_number_of_synapses\nfrom helpers import get_weight\nfrom helpers import plot_raster\nfrom helpers import fire_rate\nfrom helpers import boxplot\nfrom helpers import compute_DC\n\n\nclass Network:\n    \"\"\" Handles the setup of the network parameters and\n    provides functions to connect the network and devices.\n\n    Arguments\n    ---------\n    sim_dict\n        dictionary containing all parameters specific to the simulation\n        such as the directory the data is stored in and the seeds\n        (see: sim_params.py)\n    net_dict\n         dictionary containing all parameters specific to the neurons\n         and the network (see: network_params.py)\n\n    Keyword Arguments\n    -----------------\n    stim_dict\n        dictionary containing all parameter specific to the stimulus\n        (see: stimulus_params.py)\n\n    \"\"\"\n    def __init__(self, sim_dict, net_dict, stim_dict=None):\n        self.sim_dict = sim_dict\n        self.net_dict = net_dict\n        if stim_dict is not None:\n            self.stim_dict = stim_dict\n        else:\n            self.stim_dict = None\n        self.data_path = sim_dict['data_path']\n        if nest.Rank() == 0:\n            if os.path.isdir(self.sim_dict['data_path']):\n                print('data directory already exists')\n            else:\n                os.mkdir(self.sim_dict['data_path'])\n                print('data directory created')\n            print('Data will be written to %s' % self.data_path)\n\n    def setup_nest(self):\n        \"\"\" Hands parameters to the NEST-kernel.\n\n        Resets the NEST-kernel and passes parameters to it.\n        The number of seeds for the NEST-kernel is computed, based on the\n        total number of MPI processes and threads of each.\n\n        \"\"\"\n        nest.ResetKernel()\n        master_seed = self.sim_dict['master_seed']\n        if nest.Rank() == 0:\n            print('Master seed: %i ' % master_seed)\n        nest.SetKernelStatus(\n            {'local_num_threads': self.sim_dict['local_num_threads']}\n            )\n        N_tp = nest.GetKernelStatus(['total_num_virtual_procs'])[0]\n        if nest.Rank() == 0:\n            print('Number of total processes: %i' % N_tp)\n        rng_seeds = list(\n            range(\n                master_seed + 1 + N_tp,\n                master_seed + 1 + (2 * N_tp)\n                )\n            )\n        grng_seed = master_seed + N_tp\n        if nest.Rank() == 0:\n            print(\n                'Seeds for random number generators of virtual processes: %r'\n                % rng_seeds\n                )\n            print('Global random number generator seed: %i' % grng_seed)\n        self.pyrngs = [np.random.RandomState(s) for s in list(range(\n            master_seed, master_seed + N_tp))]\n        self.sim_resolution = self.sim_dict['sim_resolution']\n        kernel_dict = {\n            'resolution': self.sim_resolution,\n            'grng_seed': grng_seed,\n            'rng_seeds': rng_seeds,\n            'overwrite_files': self.sim_dict['overwrite_files'],\n            'print_time': self.sim_dict['print_time'],\n            }\n        nest.SetKernelStatus(kernel_dict)\n\n    def create_populations(self):\n        \"\"\" Creates the neuronal populations.\n\n        The neuronal populations are created and the parameters are assigned\n        to them. The initial membrane potential of the neurons is drawn from a\n        normal distribution. Scaling of the number of neurons and of the\n        synapses is performed. If scaling is performed extra DC input is added\n        to the neuronal populations.\n\n        \"\"\"\n        self.N_full = self.net_dict['N_full']\n        self.N_scaling = self.net_dict['N_scaling']\n        self.K_scaling = self.net_dict['K_scaling']\n        self.synapses = get_total_number_of_synapses(self.net_dict)\n        self.synapses_scaled = self.synapses * self.K_scaling\n        self.nr_neurons = self.N_full * self.N_scaling\n        self.K_ext = self.net_dict['K_ext'] * self.K_scaling\n        self.w_from_PSP = get_weight(self.net_dict['PSP_e'], self.net_dict)\n        self.weight_mat = get_weight(\n            self.net_dict['PSP_mean_matrix'], self.net_dict\n            )\n        self.weight_mat_std = self.net_dict['PSP_std_matrix']\n        self.w_ext = self.w_from_PSP\n        if self.net_dict['poisson_input']:\n            self.DC_amp_e = np.zeros(len(self.net_dict['populations']))\n        else:\n            if nest.Rank() == 0:\n                print(\n                    \"\"\"\n                    no poisson input provided\n                    calculating dc input to compensate\n                    \"\"\"\n                    )\n            self.DC_amp_e = compute_DC(self.net_dict, self.w_ext)\n\n        v0_type_options = ['original', 'optimized']\n        if self.net_dict['V0_type'] not in v0_type_options:\n            print(\n                '''\n                '{0}' is not a valid option, replacing it with '{1}'\n                Valid options are {2}\n                '''.format(self.net_dict['V0_type'],\n                           v0_type_options[0],\n                           v0_type_options)\n                )\n            self.net_dict['V0_type'] = v0_type_options[0]\n        if nest.Rank() == 0:\n            print(\n                'The number of neurons is scaled by a factor of: %.2f'\n                % self.N_scaling\n                )\n            print(\n                'The number of synapses is scaled by a factor of: %.2f'\n                % self.K_scaling\n                )\n\n        # Scaling of the synapses.\n        if self.K_scaling != 1:\n            synapses_indegree = self.synapses / (\n                self.N_full.reshape(len(self.N_full), 1) * self.N_scaling)\n            self.weight_mat, self.w_ext, self.DC_amp_e = adj_w_ext_to_K(\n                synapses_indegree, self.K_scaling, self.weight_mat,\n                self.w_from_PSP, self.DC_amp_e, self.net_dict, self.stim_dict\n                )\n\n        # Create cortical populations.\n        self.pops = []\n        pop_file = open(\n            os.path.join(self.data_path, 'population_GIDs.dat'), 'w+'\n            )\n        for i, pop in enumerate(self.net_dict['populations']):\n            population = nest.Create(\n                self.net_dict['neuron_model'], int(self.nr_neurons[i])\n                )\n            nest.SetStatus(\n                population, {\n                    'tau_syn_ex': self.net_dict['neuron_params']['tau_syn_ex'],\n                    'tau_syn_in': self.net_dict['neuron_params']['tau_syn_in'],\n                    'E_L': self.net_dict['neuron_params']['E_L'],\n                    'V_th': self.net_dict['neuron_params']['V_th'],\n                    'V_reset':  self.net_dict['neuron_params']['V_reset'],\n                    't_ref': self.net_dict['neuron_params']['t_ref'],\n                    'I_e': self.DC_amp_e[i]\n                    }\n                )\n            if self.net_dict['V0_type'] == 'optimized':\n                for thread in \\\n                        np.arange(nest.GetKernelStatus('local_num_threads')):\n                    # Using GetNodes is a work-around until NEST 3.0 is\n                    # released. It will issue a deprecation warning.\n                    local_nodes = nest.GetNodes(\n                        [0], {\n                            'model': self.net_dict['neuron_model'],\n                            'thread': thread\n                        }, local_only=True\n                    )[0]\n                    vp = nest.GetStatus(local_nodes)[0]['vp']\n                    # vp is the same for all local nodes on the same thread\n                    local_pop = list(set(local_nodes).intersection(population))\n                    nest.SetStatus(\n                        local_pop, 'V_m', self.pyrngs[vp].normal(\n                            self.net_dict\n                            ['neuron_params']['V0_mean']['optimized'][i],\n                            self.net_dict\n                            ['neuron_params']['V0_sd']['optimized'][i],\n                            len(local_pop))\n                    )\n            self.pops.append(population)\n            pop_file.write('%d  %d \\n' % (population[0], population[-1]))\n        pop_file.close()\n        if self.net_dict['V0_type'] == 'original':\n            for thread in np.arange(nest.GetKernelStatus('local_num_threads')):\n                local_nodes = nest.GetNodes(\n                    [0], {\n                        'model': self.net_dict['neuron_model'],\n                        'thread': thread\n                        }, local_only=True\n                    )[0]\n                vp = nest.GetStatus(local_nodes)[0]['vp']\n                nest.SetStatus(\n                    local_nodes, 'V_m', self.pyrngs[vp].normal(\n                        self.net_dict['neuron_params']['V0_mean']['original'],\n                        self.net_dict['neuron_params']['V0_sd']['original'],\n                        len(local_nodes))\n                        )\n\n    def create_devices(self):\n        \"\"\" Creates the recording devices.\n\n        Only devices which are given in net_dict['rec_dev'] are created.\n\n        \"\"\"\n        self.spike_detector = []\n        self.voltmeter = []\n        for i, pop in enumerate(self.pops):\n            if 'spike_detector' in self.net_dict['rec_dev']:\n                recdict = {\n                    'withgid': True,\n                    'withtime': True,\n                    'to_memory': False,\n                    'to_file': True,\n                    'label': os.path.join(self.data_path, 'spike_detector')\n                    }\n                dummy = nest.Create('spike_detector', params=recdict)\n                self.spike_detector.append(dummy)\n            if 'voltmeter' in self.net_dict['rec_dev']:\n                recdictmem = {\n                    'interval': self.sim_dict['rec_V_int'],\n                    'withgid': True,\n                    'withtime': True,\n                    'to_memory': False,\n                    'to_file': True,\n                    'label': os.path.join(self.data_path, 'voltmeter'),\n                    'record_from': ['V_m'],\n                    }\n                volt = nest.Create('voltmeter', params=recdictmem)\n                self.voltmeter.append(volt)\n\n        if 'spike_detector' in self.net_dict['rec_dev']:\n            if nest.Rank() == 0:\n                print('Spike detectors created')\n        if 'voltmeter' in self.net_dict['rec_dev']:\n            if nest.Rank() == 0:\n                print('Voltmeters created')\n\n    def create_thalamic_input(self):\n        \"\"\" This function creates the thalamic neuronal population if this\n        is specified in stimulus_params.py.\n\n        \"\"\"\n        if self.stim_dict['thalamic_input']:\n            if nest.Rank() == 0:\n                print('Thalamic input provided')\n            self.thalamic_population = nest.Create(\n                'parrot_neuron', self.stim_dict['n_thal']\n                )\n            self.thalamic_weight = get_weight(\n                self.stim_dict['PSP_th'], self.net_dict\n                )\n            self.stop_th = (\n                self.stim_dict['th_start'] + self.stim_dict['th_duration']\n                )\n            self.poisson_th = nest.Create('poisson_generator')\n            nest.SetStatus(\n                self.poisson_th, {\n                    'rate': self.stim_dict['th_rate'],\n                    'start': self.stim_dict['th_start'],\n                    'stop': self.stop_th\n                    }\n                )\n            nest.Connect(self.poisson_th, self.thalamic_population)\n            self.nr_synapses_th = synapses_th_matrix(\n                self.net_dict, self.stim_dict\n                )\n            if self.K_scaling != 1:\n                self.thalamic_weight = self.thalamic_weight / (\n                    self.K_scaling ** 0.5)\n                self.nr_synapses_th = (self.nr_synapses_th * self.K_scaling)\n        else:\n            if nest.Rank() == 0:\n                print('Thalamic input not provided')\n\n    def create_poisson(self):\n        \"\"\" Creates the Poisson generators.\n\n        If Poissonian input is provided, the Poissonian generators are created\n        and the parameters needed are passed to the Poissonian generator.\n\n        \"\"\"\n        if self.net_dict['poisson_input']:\n            if nest.Rank() == 0:\n                print('Poisson background input created')\n            rate_ext = self.net_dict['bg_rate'] * self.K_ext\n            self.poisson = []\n            for i, target_pop in enumerate(self.pops):\n                poisson = nest.Create('poisson_generator')\n                nest.SetStatus(poisson, {'rate': rate_ext[i]})\n                self.poisson.append(poisson)\n\n    def create_dc_generator(self):\n        \"\"\" Creates a DC input generator.\n\n        If DC input is provided, the DC generators are created and the\n        necessary parameters are passed to them.\n\n        \"\"\"\n        if self.stim_dict['dc_input']:\n            if nest.Rank() == 0:\n                print('DC generator created')\n            dc_amp_stim = self.net_dict['K_ext'] * self.stim_dict['dc_amp']\n            self.dc = []\n            if nest.Rank() == 0:\n                print('DC_amp_stim', dc_amp_stim)\n            for i, target_pop in enumerate(self.pops):\n                dc = nest.Create(\n                    'dc_generator', params={\n                        'amplitude': dc_amp_stim[i],\n                        'start': self.stim_dict['dc_start'],\n                        'stop': (\n                            self.stim_dict['dc_start'] +\n                            self.stim_dict['dc_dur']\n                            )\n                        }\n                    )\n                self.dc.append(dc)\n\n    def create_connections(self):\n        \"\"\" Creates the recurrent connections.\n\n        The recurrent connections between the neuronal populations are created.\n\n        \"\"\"\n        if nest.Rank() == 0:\n            print('Recurrent connections are established')\n        mean_delays = self.net_dict['mean_delay_matrix']\n        std_delays = self.net_dict['std_delay_matrix']\n        for i, target_pop in enumerate(self.pops):\n            for j, source_pop in enumerate(self.pops):\n                synapse_nr = int(self.synapses_scaled[i][j])\n                if synapse_nr >= 0.:\n                    weight = self.weight_mat[i][j]\n                    w_sd = abs(weight * self.weight_mat_std[i][j])\n                    conn_dict_rec = {\n                        'rule': 'fixed_total_number', 'N': synapse_nr\n                        }\n                    syn_dict = {\n                        'model': 'static_synapse',\n                        'weight': {\n                            'distribution': 'normal_clipped', 'mu': weight,\n                            'sigma': w_sd\n                            },\n                        'delay': {\n                            'distribution': 'normal_clipped',\n                            'mu': mean_delays[i][j], 'sigma': std_delays[i][j],\n                            'low': self.sim_resolution\n                            }\n                        }\n                    if weight < 0:\n                        syn_dict['weight']['high'] = 0.0\n                    else:\n                        syn_dict['weight']['low'] = 0.0\n                    nest.Connect(\n                        source_pop, target_pop,\n                        conn_spec=conn_dict_rec,\n                        syn_spec=syn_dict\n                        )\n\n    def connect_poisson(self):\n        \"\"\" Connects the Poisson generators to the microcircuit.\"\"\"\n        if nest.Rank() == 0:\n            print('Poisson background input is connected')\n        for i, target_pop in enumerate(self.pops):\n            conn_dict_poisson = {'rule': 'all_to_all'}\n            syn_dict_poisson = {\n                'model': 'static_synapse',\n                'weight': self.w_ext,\n                'delay': self.net_dict['poisson_delay']\n                }\n            nest.Connect(\n                self.poisson[i], target_pop,\n                conn_spec=conn_dict_poisson,\n                syn_spec=syn_dict_poisson\n                )\n\n    def connect_thalamus(self):\n        \"\"\" Connects the thalamic population to the microcircuit.\"\"\"\n        if nest.Rank() == 0:\n            print('Thalamus connection established')\n        for i, target_pop in enumerate(self.pops):\n            conn_dict_th = {\n                'rule': 'fixed_total_number',\n                'N': int(self.nr_synapses_th[i])\n                }\n            syn_dict_th = {\n                'weight': {\n                    'distribution': 'normal_clipped',\n                    'mu': self.thalamic_weight,\n                    'sigma': (\n                        self.thalamic_weight * self.net_dict['PSP_sd']\n                        ),\n                    'low': 0.0\n                    },\n                'delay': {\n                    'distribution': 'normal_clipped',\n                    'mu': self.stim_dict['delay_th'][i],\n                    'sigma': self.stim_dict['delay_th_sd'][i],\n                    'low': self.sim_resolution\n                    }\n                }\n            nest.Connect(\n                self.thalamic_population, target_pop,\n                conn_spec=conn_dict_th, syn_spec=syn_dict_th\n                )\n\n    def connect_dc_generator(self):\n        \"\"\" Connects the DC generator to the microcircuit.\"\"\"\n        if nest.Rank() == 0:\n            print('DC Generator connection established')\n        for i, target_pop in enumerate(self.pops):\n            if self.stim_dict['dc_input']:\n                nest.Connect(self.dc[i], target_pop)\n\n    def connect_devices(self):\n        \"\"\" Connects the recording devices to the microcircuit.\"\"\"\n        if nest.Rank() == 0:\n            if ('spike_detector' in self.net_dict['rec_dev'] and\n                    'voltmeter' not in self.net_dict['rec_dev']):\n                print('Spike detector connected')\n            elif ('spike_detector' not in self.net_dict['rec_dev'] and\n                    'voltmeter' in self.net_dict['rec_dev']):\n                print('Voltmeter connected')\n            elif ('spike_detector' in self.net_dict['rec_dev'] and\n                    'voltmeter' in self.net_dict['rec_dev']):\n                print('Spike detector and voltmeter connected')\n            else:\n                print('no recording devices connected')\n        for i, target_pop in enumerate(self.pops):\n            if 'voltmeter' in self.net_dict['rec_dev']:\n                nest.Connect(self.voltmeter[i], target_pop)\n            if 'spike_detector' in self.net_dict['rec_dev']:\n                nest.Connect(target_pop, self.spike_detector[i])\n\n    def setup(self):\n        \"\"\" Execute subfunctions of the network.\n\n        This function executes several subfunctions to create neuronal\n        populations, devices and inputs, connects the populations with\n        each other and with devices and input nodes.\n\n        \"\"\"\n        self.setup_nest()\n        self.create_populations()\n        self.create_devices()\n        self.create_thalamic_input()\n        self.create_poisson()\n        self.create_dc_generator()\n        self.create_connections()\n        if self.net_dict['poisson_input']:\n            self.connect_poisson()\n        if self.stim_dict['thalamic_input']:\n            self.connect_thalamus()\n        if self.stim_dict['dc_input']:\n            self.connect_dc_generator()\n        self.connect_devices()\n\n    def simulate(self):\n        \"\"\" Simulates the microcircuit.\"\"\"\n        nest.Simulate(self.sim_dict['t_sim'])\n\n    def evaluate(self, raster_plot_time_idx, fire_rate_time_idx):\n        \"\"\" Displays output of the simulation.\n\n        Calculates the firing rate of each population,\n        creates a spike raster plot and a box plot of the\n        firing rates.\n\n        \"\"\"\n        if nest.Rank() == 0:\n            print(\n                'Interval to compute firing rates: %s ms'\n                % np.array2string(fire_rate_time_idx)\n                )\n            fire_rate(\n                self.data_path, 'spike_detector',\n                fire_rate_time_idx[0], fire_rate_time_idx[1]\n                )\n            print(\n                'Interval to plot spikes: %s ms'\n                % np.array2string(raster_plot_time_idx)\n                )\n            plot_raster(\n                self.data_path, 'spike_detector',\n                raster_plot_time_idx[0], raster_plot_time_idx[1]\n                )\n            boxplot(self.net_dict, self.data_path)"
      ]
    }
  ],
  "metadata": {
    "kernelspec": {
      "display_name": "Python 3",
      "language": "python",
      "name": "python3"
    },
    "language_info": {
      "codemirror_mode": {
        "name": "ipython",
        "version": 3
      },
      "file_extension": ".py",
      "mimetype": "text/x-python",
      "name": "python",
      "nbconvert_exporter": "python",
      "pygments_lexer": "ipython3",
      "version": "3.6.8"
    }
  },
  "nbformat": 4,
  "nbformat_minor": 0
}